import inspect
import time

import httpx

from nacos.api.base import BaseNacosAPI
from nacos.api.models import AccessToken
from nacos.errors import _default_response_to_exception


class BaseNacosClient:

    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls)
        api_endpoints = inspect.getmembers(self, lambda x: isinstance(x, BaseNacosAPI))
        for name, api in api_endpoints:
            api_cls = type(api)
            api = api_cls(self)
            setattr(self, name, api)
        return self

    def __init__(self,
                 server_addresses: str,
                 username: str,
                 password: str,
                 namespace_id: str = None):
        self.server_addresses = server_addresses
        self.username = username
        self.password = password
        self.namespace_id = namespace_id

        self.access_token = None
        self.access_token_timestamp = None
        self.access_token_ttl = None

        self.request = httpx.Client(
            base_url=self.server_addresses,
            headers={"Content-Type": "application/x-www-form-urlencoded"},
            event_hooks={
                "response": [_default_response_to_exception],
            },
            timeout=10
        )

        self._fetch_access_token()

    def _fetch_access_token(self):
        """ Fetches the access token """
        now = int(time.time())
        if self.access_token is None or self.access_token_timestamp + self.access_token_ttl * 0.9 < now:
            res = self.request.post('/nacos/v1/auth/login',
                                    data={'username': self.username, 'password': self.password})
            data = AccessToken.parse_raw(res.content)

            self.access_token = data.accessToken
            self.access_token_timestamp = now
            self.access_token_ttl = data.tokenTtl

        return self.access_token

    def _do_request(self, method, url, **kwargs):
        access_token = self._fetch_access_token()
        return self.request.request(method, f'{url}?accessToken={access_token}', **kwargs)



from nacos.api.base import BaseNacosAPI


class NacosConfig(BaseNacosAPI):

    def get(self,
            group: str,
            dataId: str,
            *,
            namespace_id: str = 'public',
            tag: str = ''):
        """获取指定配置

        Parameters
        ----------
        group
            配置分组名
        dataId
            配置名
        namespace_id, optional
            命名空间, by default 'public'
        tag, optional
            标签, by default ''

        Returns
        -------
            _description_
        """
        params = {
            'dataId': dataId,
            'group': group,
            'namespaceId': namespace_id,
            'tag': tag
        }
        res = self._get('/nacos/v2/cs/config', params=params)
        return res

    def create(self,
               group: str,
               data_id: str,
               content: str,
               *,
               namespace_id: str = 'public',
               tag: str = '',
               app_name: str = '',
               src_user: str = '',
               config_tags: str = '',
               desc: str = '',
               effect: str = '',
               type: str = '',
               use: str = '',
               schema: str = '',
               ):
        """发布指定配置
        当配置已存在时，则对配置进行更新

        Parameters
        ----------
        group
            配置组名
        data_id
            配置名
        content
            配置内容
        namespace_id, optional
            命名空间, by default 'public'
        tag, optional
            标签, by default ''
        app_name, optional
            应用名, by default ''
        src_user, optional
            源用户, by default ''
        config_tags, optional
            配置标签列表,可多个,逗号分隔, by default ''
        desc, optional
            配置描述, by default ''
        effect, optional
            by default ''
        type, optional
            配置类型, by default ''
        use, optional
            by default ''
        schema, optional
            by default ''
        """
        data = {
            'group': group,
            'dataId': data_id,
            'content': content,
            'namespaceId': namespace_id,
            'tag': tag,
            'configTags': config_tags,
            'desc': desc,
            'effect': effect,
            'type': type,
            'appName': app_name,
            'srcUser': src_user,
            'use': use,
            'schema': schema,
        }
        return self._post('/nacos/v2/cs/config', data=data)

    def delete(self,
               group: str,
               dataId: str,
               *,
               namespace_id: str = 'public',
               tag: str = ''):
        """删除指定配置

        Parameters
        ----------
        group
            配置分组名
        dataId
            配置名
        namespace_id, optional
            命名空间, by default 'public'
        tag, optional
            标签, by default ''

        """
        params = {
            'dataId': dataId,
            'group': group,
            'namespaceId': namespace_id,
            'tag': tag
        }
        return self._delete('/nacos/v2/cs/config', params=params)

    def get_history_list(self,
                         group: str,
                         dataId: str,
                         *,
                         namespace_id: str = 'public',
                         page: int = 1,
                         page_size: int = 100):
        """获取指定配置的历史版本列表

        Parameters
        ----------
        group
            配置分组名
        dataId
            配置名
        namespace_id, optional
            命名空间, by default 'public'
        page, optional
            当前页, by default 1
        page_size, optional
            页条目数, 最大为500, by default 100
        """
        params = {
            'group': group,
            'dataId': dataId,
            'namespaceId': namespace_id,
            'pageNo': page,
            'pageSize': page_size
        }
        return self._get('/nacos/v2/cs/history/list', params=params)

    def get_history(self,
                    group: str,
                    dataId: str,
                    nid: int,
                    *,
                    namespace_id: str = 'public'):
        """获取指定版本的历史配置

        Parameters
        ----------
        group
            配置分组名
        dataId
            配置名
        nid
            历史配置id
        namespace_id, optional
            命名空间, by default 'public'
        """
        params = {
            'group': group,
            'dataId': dataId,
            'nid': nid,
            'namespaceId': namespace_id
        }
        return self._get('/nacos/v2/cs/history', params=params)

    def get_history_previous(self,
                             group: str,
                             dataId: str,
                             id: int,
                             *,
                             namespace_id: str = 'public'):
        """获取指定版本的历史配置

        Parameters
        ----------
        group
            配置分组名
        dataId
            配置名
        nid
            历史配置id
        namespace_id, optional
            命名空间, by default 'public'
        """
        params = {
            'group': group,
            'dataId': dataId,
            'id': id,
            'namespaceId': namespace_id
        }
        return self._get('/nacos/v2/cs/history/previous', params=params)

    def get_by_namespace(self, namespace_id: str):
        """获取指定命名空间下的配置信息列表

        Parameters
        ----------
        namespace_id
            命名空间
        """
        return self._get('/nacos/v2/cs/history/configs', params={'namespaceId': namespace_id})



import typing
from datetime import datetime
from enum import Enum

from nacos.models import BaseModel


class AccessToken(BaseModel):
    accessToken: str
    tokenTtl: int


class NacosBaseModel(BaseModel):
    code: int
    message: typing.Optional[str]
    data: typing.Union[dict, list, bool]


class MetaDataModel(BaseModel):
    value: str


class InstanceModel(BaseModel):
    '''Instance'''
    serviceName: typing.Optional[str]
    service: typing.Union[str, None]
    ip: str
    port: int
    clusterName: str
    weight: float
    healthy: bool
    enabled: typing.Optional[bool]
    ephemeral: typing.Optional[bool]
    instanceId: typing.Optional[str]
    instanceHeartBeatInterval: typing.Optional[int]
    instanceIdGenerator: typing.Optional[str]
    instanceHeartBeatTimeOut: typing.Optional[int]
    ipDeleteTimeout: typing.Optional[int]
    metadata: typing.Union[MetaDataModel, dict]


class InstanceListModel(BaseModel):
    '''Instance list'''
    name: str
    groupName: str
    clusters: str
    cacheMillis: int
    lastRefTime: int
    checksum: str
    allIPs: bool
    reachProtectionThreshold: bool
    valid: bool
    hosts: typing.List[InstanceModel]


class InstanceSendHeartBeatModel(BaseModel):
    ''' Instance send heartbeat'''
    clientBeatInterval: int
    code: int
    lightBeatEnabled: bool


class ServiceListDataModel(BaseModel):
    count: int
    services: typing.List[str]


class ServiceListModel(NacosBaseModel):
    data: typing.Union[ServiceListDataModel, dict]


class ServiceInfoModel(BaseModel):
    namespaceId: str
    groupName: str
    name: str
    protectThreshold: float
    metadata: dict
    selector: dict
    clusters: typing.List[dict]



import json
import logging
import time

import makepath

import config
from nacos import NacosClient

logging.basicConfig(level=logging.INFO, format='%(asctime)-15s %(levelname)s [%(pathname)s:%(lineno)d] %(message)s')
logger = logging.getLogger(__name__)

namespace_id = '26b052f1-f8c9-4f31-8900-affd3c4c9e1c'
service_name = ''
group_name = 'public'
ip = '127.0.0.1'
port = 5000

nc = NacosClient(**config.NACOS_CONFIG)

while True:
    print(nc.get_random_instance('monitor'))
    time.sleep(2)

# print(nc.service.get_by_namespace(namespace_id, with_instances=True).content)
# print(nc.service.get('hsdp-schedule', namespace_id=namespace_id, group_name=group_name))



from nacos.api.base import BaseNacosAPI

from .models import ServiceInfoModel, ServiceListModel


class NacosService(BaseNacosAPI):

    def get(self,
            service_name: str,
            *,
            namespace_id: str = 'public',
            group_name: str = 'DEFAULT_GROUP'):
        """查询某个具体服务的详情信息

        Parameters
        ----------
        service_name
            服务名
        namespace_id, optional
            命名空间Id, by default 'public'
        group_name, optional
            分组名, by default 'DEFAULT_GROUP'

        Example
        ----------
            from nacos.client import NacosClient

            client = NacosClient('server_addresses', 'namespace', 'username', 'password')
            response = client.service.get('service_name')
        """

        params = {
            'serviceName': f'{service_name}',
            'namespaceId': namespace_id,
            'groupName': group_name
        }
        res = self._get('/nacos/v1/ns/service', params=params)
        return ServiceInfoModel.parse_raw(res.content)

    def get_by_namespace(self,
                         namespace_id,
                         *,
                         service_name: str = None,
                         group_name: str = None,
                         has_ip_count: bool = True,
                         with_instances: bool = False,
                         page: int = 1,
                         page_size: int = 20):
        """ 查询指定命名空间下的服务列表

        Parameters
        ----------
        namespace_id
            命名空间
        service_name, optional
            服务名, by default None
        group_name, optional
            分组名, by default None
        has_ip_count, optional
            是否包含ip统计, by default True
        with_instances, optional
            是否返回示例列表, by default False
        page, optional
            当前页, by default 1
        page_size, optional
            页条目数, by default 20

        """
        data = {
            'namespaceId': namespace_id,
            'serviceNameParam': service_name,
            'groupNameParam': group_name,
            'withInstances': with_instances,
            'hasIpCount': has_ip_count,
            'pageNo': page,
            'pageSize': page_size
        }
        return self._get('/nacos/v1/ns/catalog/services', params=data)

    def query(self,
              selector: str,
              *,
              namespace_id: str = 'public',
              group_name: str = 'DEFAULT_GROUP',
              page: int = 1,
              page_size: int = 20):
        """查询符合条件的服务列表

        Parameters
        ----------
        selector
            访问策略
        namespace_id, optional
            命名空间Id, by default 'public'
        group_name, optional
            分组名, by default 'DEFAULT_GROUP'
        page, optional
            当前页, by default 1
        page_size, optional
            页条目数, by default 20

        Returns
        -------
            _description_
        """

        params = {
            'selector': selector,
            'namespaceId': namespace_id,
            'groupName': group_name,
            'pageNo': page,
            'pageSize': page_size}
        res = self._get('/nacos/v2/ns/service/list', params=params)
        return ServiceListModel.parse_raw(res.content)

    def create(self,
               service_name: str,
               *,
               namespace_id: str = 'public',
               group_name: str = 'DEFAULT_GROUP',
               metadata: str = '{}',
               protect_threshold: float = 0,
               ephemeral: bool = False,
               selector: str = ''):
        """注册实例

        Parameters
        ----------
        service_name
            服务名
        namespace_id, optional
            命名空间Id, by default 'public'
        group_name, optional
            分组名, by default 'DEFAULT_GROUP'
        metadata, optional
            实例元数据, by default "{}"
        ephemeral, optional
            是否为临时实例, by default False
        protect_threshold, optional
            保护阈值, by default 0
        selector, optional
            访问策略, by default ''
        """

        data = {
            'serviceName': f'{group_name}@@{service_name}',
            'namespaceId': namespace_id,
            'groupName': group_name,
            'selector': selector,
            'protectThreshold': protect_threshold,
            'metadata': metadata,
            'ephemeral': ephemeral
        }
        return self._post('/nacos/v2/ns/service', data=data)

    def update(self,
               service_name: str,
               *,
               namespace_id: str = 'public',
               group_name: str = 'DEFAULT_GROUP',
               metadata: str = "{}",
               protect_threshold: float = 0,
               selector: str = ''):
        """更新指定服务
        服务不存在时会报错

        Parameters
        ----------
        service_name
            服务名
        namespace_id, optional
            命名空间Id, by default 'public'
        group_name, optional
            分组名, by default 'DEFAULT_GROUP'
        metadata, optional
            实例元数据, by default "{}"
        protect_threshold, optional
            保护阈值, by default 0
        selector, optional
            访问策略, by default ''
        """

        data = {
            'serviceName': f'{group_name}@@{service_name}',
            'namespaceId': namespace_id,
            'groupName': group_name,
            'metadata': metadata,
            'protectThreshold': protect_threshold,
            'selector': selector
        }
        return self._post('/nacos/v2/ns/service', data=data)

    def delete(self,
               service_name: str,
               *,
               namespace_id: str = 'public',
               group_name: str = 'DEFAULT_GROUP'):
        """删除指定服务
        服务不存在时会报错，且服务还存在实例时会删除失败

        Parameters
        ----------
        service_name
            服务名
        namespace_id, optional
            命名空间Id, by default 'public'
        group_name, optional
            分组名, by default 'DEFAULT_GROUP'
        """

        data = {
            'serviceName': f'{group_name}@@{service_name}',
            'namespaceId': namespace_id,
            'groupName': group_name,
        }
        return self._delete('/nacos/v2/ns/service', data=data)



from nacos.api.base import BaseNacosAPI
from nacos.api.models import InstanceListModel, InstanceModel, InstanceSendHeartBeatModel


class NacosInstance(BaseNacosAPI):

    def get(self,
            service_name: str,
            ip: str,
            port: int,
            *,
            namespace_id: str = 'public',
            group_name: str = 'DEFAULT_GROUP',
            cluster_name: str = 'DEFAULT'):
        """查询某个具体实例的详情信息

        Parameters
        ----------
        service_name
            服务名
        ip
            IP地址
        port
            端口号
        namespace_id, optional
            命名空间Id, by default 'public'
        group_name, optional
            分组名, by default 'DEFAULT_GROUP'
        cluster_name, optional
            集群名称, by default 'DEFAULT'

        Example
        ----------
            from nacos.client import NacosClient

            client = NacosClient('server_addresses', 'namespace', 'username', 'password')
            response = client.instance.get('service_name', '127.0.0.1', 8000)
        """

        params = {
            'serviceName': f'{group_name}@@{service_name}',
            'ip': ip,
            'port': port,
            'namespaceId': namespace_id,
            'groupName': group_name,
            'clusterName': cluster_name
        }
        res = self._get('/nacos/v2/ns/instance', params=params)
        return InstanceModel.parse_raw(res.content)

    def query(self,
              service_name: str,
              *,
              ip: str = '',
              port: int = 0,
              namespace_id: str = 'public',
              group_name: str = 'DEFAULT_GROUP',
              cluster_name: str = 'DEFAULT',
              healthy_only: bool = False,
              app: str = ''):
        """查询指定服务下的实例详情信息列表

        Parameters
        ----------
        service_name
            服务名
        ip
            IP地址
        port
            端口号
        namespace_id, optional
            命名空间Id, by default 'public'
        group_name, optional
            分组名, by default 'DEFAULT_GROUP'
        cluster_name, optional
            集群名称, by default 'DEFAULT'
        healthy_only, optional
            是否只获取健康实例, by default False
        app, optional
            应用名, by default ''

        Returns
        -------
            InstanceListModel
        """

        params = {
            'serviceName': f'{group_name}@@{service_name}',
            'ip': ip,
            'port': port,
            'namespaceId': namespace_id,
            'groupName': group_name,
            'clusterName': cluster_name,
            'healthyOnly': healthy_only,
            'app': app
        }
        res = self._get('/nacos/v2/ns/instance/list', params=params)
        return InstanceListModel.parse_raw(res.content)

    def create(self,
               service_name: str,
               ip: str,
               port: int,
               *,
               namespace_id: str = 'public',
               group_name: str = 'DEFAULT_GROUP',
               cluster_name: str = 'DEFAULT',
               healthy: bool = True,
               weight: float = 1.0,
               enabled: bool = True,
               metadata: str = '{}',
               ephemeral: bool = False):
        """注册实例

        Parameters
        ----------
        service_name
            服务名
        ip
            IP地址
        port
            端口号
        namespace_id, optional
            命名空间Id, by default 'public'
        group_name, optional
            分组名, by default 'DEFAULT_GROUP'
        cluster_name, optional
            集群名称, by default 'DEFAULT'
        healthy, optional
            是否只查找健康实例, by default True
        weight, optional
            实例权重, by default 1.0
        enabled, optional
            是否可用, by default True
        metadata, optional
            实例元数据, by default "{}"
        ephemeral, optional
            是否为临时实例, by default False
        """

        data = {
            'serviceName': f'{group_name}@@{service_name}',
            'ip': ip,
            'port': port,
            'namespaceId': namespace_id,
            'groupName': group_name,
            'clusterName': cluster_name,
            'healthy': healthy,
            'weight': weight,
            'enabled': enabled,
            'metadata': metadata,
            'ephemeral': ephemeral
        }
        return self._post('/nacos/v2/ns/instance', data=data)

    def update(self,
               service_name: str,
               ip: str,
               port: int,
               *,
               namespace_id: str = 'public',
               group_name: str = 'DEFAULT_GROUP',
               cluster_name: str = 'DEFAULT',
               healthy: bool = True,
               weight: float = 1.0,
               enabled: bool = True,
               metadata: str = "{}",
               ephemeral: bool = False):
        """修改指定实例

        Parameters
        ----------
        service_name
            服务名
        ip
            IP地址
        port
            端口号
        namespace_id, optional
            命名空间Id, by default 'public'
        group_name, optional
            分组名, by default 'DEFAULT_GROUP'
        cluster_name, optional
            集群名称, by default 'DEFAULT'
        healthy, optional
            是否只查找健康实例, by default True
        weight, optional
            实例权重, by default 1.0
        enabled, optional
            是否可用, by default True
        metadata, optional
            实例元数据, by default "{}"
        ephemeral, optional
            是否为临时实例, by default False
        """

        data = {
            'serviceName': f'{group_name}@@{service_name}',
            'ip': ip,
            'port': port,
            'namespaceId': namespace_id,
            'groupName': group_name,
            'clusterName': cluster_name,
            'healthy': healthy,
            'weight': weight,
            'enabled': enabled,
            'metadata': metadata,
            'ephemeral': ephemeral
        }
        return self._put('/nacos/v2/ns/instance', data=data)

    def update_health(self,
                      service_name: str,
                      ip: str,
                      port: int,
                      healthy: bool,
                      *,
                      namespace_id: str = 'public',
                      group_name: str = 'DEFAULT_GROUP',
                      cluster_name: str = 'DEFAULT'):
        """更新实例的健康状态

        Parameters
        ----------
        service_name
            服务名
        ip
            IP地址
        port
            端口号
        healthy
            是否健康
        namespace_id, optional
            命名空间Id, by default 'public'
        group_name, optional
            分组名, by default 'DEFAULT_GROUP'
        cluster_name, optional
            集群名称, by default 'DEFAULT'
        """

        data = {
            'serviceName': f'{group_name}@@{service_name}',
            'ip': ip,
            'port': port,
            'healthy': healthy,
            'namespaceId': namespace_id,
            'groupName': group_name,
            'clusterName': cluster_name
        }
        return self._put('/nacos/v2/ns/health/instance', data=data)

    def delete(self,
               service_name: str,
               ip: str,
               port: int,
               *,
               namespace_id: str = 'public',
               group_name: str = 'DEFAULT_GROUP',
               cluster_name: str = 'DEFAULT',
               healthy: bool = True,
               weight: float = 1.0,
               enabled: bool = True,
               metadata: str = "{}",
               ephemeral: bool = False):
        """注销指定实例

        Parameters
        ----------
        service_name
            服务名
        ip
            IP地址
        port
            端口号
        namespace_id, optional
            命名空间Id, by default 'public'
        group_name, optional
            分组名, by default 'DEFAULT_GROUP'
        cluster_name, optional
            集群名称, by default 'DEFAULT'
        healthy, optional
            是否只查找健康实例, by default True
        weight, optional
            实例权重, by default 1.0
        enabled, optional
            是否可用, by default True
        metadata, optional
            实例元数据, by default "{}"
        ephemeral, optional
            是否为临时实例, by default False
        """

        data = {
            'serviceName': f'{group_name}@@{service_name}',
            'ip': ip,
            'port': port,
            'namespaceId': namespace_id,
            'groupName': group_name,
            'clusterName': cluster_name,
            'healthy': healthy,
            'weight': weight,
            'enabled': enabled,
            'metadata': metadata,
            'ephemeral': ephemeral
        }
        return self._delete('/nacos/v2/ns/instance', data=data)

    def batch_update_metadata(self,
                              service_name: str,
                              metadata: str,
                              *,
                              namespace_id: str = 'public',
                              consistency_type: str = '',
                              instances: str = '[]',
                              group_name: str = 'DEFAULT_GROUP'):
        """批量更新实例的元数据

        Parameters
        ----------
        service_name
            服务名
        metadata
            实例元数据
        namespace_id, optional
            命名空间Id, by default 'public'
        consistency_type, optional
            持久化类型, by default ''
        instances, optional
            需要更新的实例列表, by default '[]'
            通过ip+port+ephemeral+cluster定位到某一实例,为空则表示更新指定服务下所有实例的元数据
        group_name, optional
            分组名, by default 'DEFAULT_GROUP'
        """
        data = {
            'serviceName': f'{group_name}@@{service_name}',
            'namespaceId': namespace_id,
            'groupName': group_name,
            'metadata': metadata,
            'consistencyType': consistency_type,
            'instances': instances
        }
        return self._put('/nacos/v2/ns/instance/metadata/batch', data=data)

    def batch_delete_metadata(self,
                              service_name: str,
                              metadata: str,
                              *,
                              namespace_id: str = 'public',
                              consistency_type: str = '',
                              instances: str = '[]',
                              group_name: str = 'DEFAULT_GROUP'):
        """批量删除实例的元数据

        Parameters
        ----------
        service_name
            服务名
        metadata
            实例元数据
        namespace_id, optional
            命名空间Id, by default 'public'
        consistency_type, optional
            持久化类型, by default ''
        instances, optional
            需要更新的实例列表, by default '[]'
            通过ip+port+ephemeral+cluster定位到某一实例,为空则表示更新指定服务下所有实例的元数据
        group_name, optional
            分组名, by default 'DEFAULT_GROUP'
        """
        data = {
            'serviceName': f'{group_name}@@{service_name}',
            'namespaceId': namespace_id,
            'groupName': group_name,
            'metadata': metadata,
            'consistencyType': consistency_type,
            'instances': instances
        }
        return self._delete('/nacos/v2/ns/instance/metadata/batch', data=data)

    def send_heartbeat(self,
                       service_name: str,
                       ip: str,
                       port: int,
                       beat: str,
                       *,
                       namespace_id: str = 'public',
                       group_name: str = 'DEFAULT_GROUP',
                       ephemeral: bool = False):
        """发送心跳

        Parameters
        ----------
        service_name
            服务名
        ip
            IP地址
        port
            端口号
        beat
            心跳信息
        namespace_id, optional
            命名空间Id, by default 'public'
        group_name, optional
            分组名, by default 'DEFAULT_GROUP'
        ephemeral, optional
            是否为临时实例, by default False
        """
        data = {
            'serviceName': f'{group_name}@@{service_name}',
            'namespaceId': namespace_id,
            'groupName': group_name,
            'ip': ip,
            'port': port,
            'beat': beat,
            'ephemeral': ephemeral
        }
        res = self._put('/nacos/v2/ns/instance/beat', data=data)
        return InstanceSendHeartBeatModel.parse_raw(res.content)



# ruff: noqa: N818
import functools
import http
import typing

import httpx


class BaseError(Exception):
    """ 基异常 """

    pass


class RequestError(BaseError):
    """请求异常"""

    def __init__(self, message: str, *, request: httpx.Request):
        """__init__

        Parameters
        ----------
        message :
            异常信息
        request :
            请求对象
        """
        super().__init__(message)
        self.request = request


class Timeout(RequestError):
    """请求超时异常"""

    pass


class ResponseError(BaseError):
    """响应异常"""

    def __init__(
        self, message: str, *, request: httpx.Request, response: httpx.Response
    ):
        """__init__

        Parameters
        ----------
        message :
            异常信息
        request :
            请求对象
        response :
            响应对象
        """
        super().__init__(message)
        self.request = request
        self.response = response


### client error ###
class ClientError(ResponseError):
    """客户端异常"""

    code = "invalid_argument"


class Canceled(ClientError):
    """请求取消"""

    code = "canceled"


class InvalidArgument(ClientError):
    """参数错误"""

    code = "invalid_argument"


class Malformed(ClientError):
    """请求格式错误"""

    code = "malformed"


class DeadlineExceeded(ClientError):
    """请求过期"""

    code = "deadline_exceeded"


class NotFound(ClientError):
    """资源不存在"""

    code = "not_found"


class BadRoute(ClientError):
    """请求路由错误"""

    code = "bad_route"


class AlreadyExists(ClientError):
    """资源已存在"""

    code = "already_exists"


class PermissionDenied(ClientError):
    """权限不足"""

    code = "permission_denied"


class Unauthenticated(ClientError):
    """未认证"""

    code = "unauthenticated"


class ResourceExhausted(ClientError):
    """资源耗尽"""

    code = "resource_exhausted"


class FailedPrecondition(ClientError):
    """请求前置条件失败"""

    code = "failed_precondition"


class Aborted(ClientError):
    """请求中止"""

    code = "aborted"


class OutOfRange(ClientError):
    """请求参数超出范围"""

    code = "out_of_range"


### server error ###
class ServerError(ResponseError):
    """服务端异常"""

    code = "internal"


class Unknown(ServerError):
    """未知异常"""

    code = "unknown"


class Unimplemented(ServerError):
    """未实现"""

    code = "unimplemented"


class Internal(ServerError):
    """内部错误"""

    code = "internal"


class Unavailable(ServerError):
    """服务不可用"""

    code = "unavailable"


class DataLoss(ServerError):
    """数据丢失"""

    code = "dataloss"


def _unify_error(fn):
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except httpx.TimeoutException as e:
            raise Timeout(str(e), request=e.request) from e
        except httpx.RequestError as e:
            raise RequestError(str(e), request=e.request) from e

    return wrapper


_STATUS_CODE_CLIENT_ERROR_MAPPING = {
    http.HTTPStatus.REQUEST_TIMEOUT: DeadlineExceeded,
    http.HTTPStatus.UNPROCESSABLE_ENTITY: InvalidArgument,
    http.HTTPStatus.NOT_FOUND: NotFound,
    http.HTTPStatus.CONFLICT: AlreadyExists,
    http.HTTPStatus.FORBIDDEN: PermissionDenied,
    http.HTTPStatus.UNAUTHORIZED: Unauthenticated,
    http.HTTPStatus.TOO_MANY_REQUESTS: ResourceExhausted,
    http.HTTPStatus.PRECONDITION_FAILED: FailedPrecondition,
}

_STATUS_CODE_SERVER_ERROR_MAPPING = {
    http.HTTPStatus.INTERNAL_SERVER_ERROR: Internal,
    http.HTTPStatus.NOT_IMPLEMENTED: Unimplemented,
    http.HTTPStatus.SERVICE_UNAVAILABLE: Unavailable,
}


def _default_response_to_exception(response: httpx.Response):
    if response.is_success:
        return
    exception: typing.Union[typing.Type[ClientError], typing.Type[ServerError]]
    if response.is_client_error:
        exception = _STATUS_CODE_CLIENT_ERROR_MAPPING.get(
            http.HTTPStatus(response.status_code), ClientError
        )
        type_ = "Client Error"
    elif response.is_server_error:
        exception = _STATUS_CODE_SERVER_ERROR_MAPPING.get(
            http.HTTPStatus(response.status_code), ServerError
        )
        type_ = "Server Error"
    else:
        raise RuntimeError(f"Unexpected response: {response}")
    message = (
        f"{type_}: '{exception.code}' for url '{response.url}'. Response: {response.read()}"
    )
    raise exception(message, request=response.request, response=response)



import json
import logging
import random
import signal
import sys
import threading
import time
from collections import defaultdict

from nacos import api
from nacos.client.base import BaseNacosClient

logger = logging.getLogger(__name__)


class NacosClient(BaseNacosClient):

    instance = api.NacosInstance()
    config = api.NacosConfig()
    service = api.NacosService()

    def __init__(self,
                 server_addresses: str,
                 username: str,
                 password: str,
                 namespace_id: str = None):
        super().__init__(server_addresses, username, password, namespace_id)

        self.service_dict = defaultdict(lambda: defaultdict(list))
        self.register_info = None

        if namespace_id:
            subscribe_thread = threading.Thread(target=self.subscribe_services)
            subscribe_thread.setDaemon(True)
            subscribe_thread.start()

    def get_random_instance(self, service_name, group_name='public', cluster_name="DEFAULT"):
        if not self.namespace_id:
            return None
        hosts = self.service_dict[f'{group_name}@@{service_name}'][cluster_name]
        if not hosts:
            # 获取instance list
            res = self.instance.query(service_name, group_name=group_name,
                                      cluster_name=cluster_name, namespace_id=self.namespace_id)
            hosts = [f'{host.ip}:{host.port}' for host in res.hosts if host.healthy and host.enabled]
            self.service_dict[f'{group_name}@@{service_name}'][cluster_name] = hosts
        return random.choice(hosts) if hosts else None

    def subscribe_services(self):
        while True:
            response = self.service.get_by_namespace(self.namespace_id, with_instances=True, page_size=999)
            for i in response.json():
                for cluster, item in i['clusterMap'].items():
                    self.service_dict[f'{i["groupName"]}@@{i["serviceName"]}'][cluster] = [
                        f"{j['ip']}:{j['port']}" for j in item['hosts'] if j['enabled']]

            time.sleep(5)

    def register(self,
                 service_name: str,
                 ip: str,
                 port: int,
                 *,
                 namespace_id: str = 'public',
                 group_name: str = 'public',
                 cluster_name: str = 'DEFAULT',
                 healthy: bool = True,
                 weight: float = 1.0,
                 enabled: bool = True,
                 metadata: str = '{}',
                 ephemeral: bool = True):
        """ 自动注册服务
        注册实例、发送心跳；服务终止时删除实例

        Parameters
        ----------
        service_name
            服务名
        ip
            IP地址
        port
            端口号
        namespace_id, optional
            命名空间Id, by default 'public'
        group_name, optional
            分组名, by default 'DEFAULT_GROUP'
        cluster_name, optional
            集群名称, by default 'DEFAULT'
        healthy, optional
            是否只查找健康实例, by default True
        weight, optional
            实例权重, by default 1.0
        enabled, optional
            是否可用, by default True
        metadata, optional
            实例元数据, by default "{}"
        ephemeral, optional
            是否为临时实例, by default True
        """

        signal.signal(signal.SIGTERM, self.delete_instance)
        signal.signal(signal.SIGINT, self.delete_instance)

        self.register_info = locals()

        try:
            # 注册实例
            logger.info(
                f'[InstanceRegister] service_name: {service_name}, ip: {ip}, port: {port}, group_name: {group_name}, namespace_id: {namespace_id}')
            result = self.instance.create(service_name, ip, port, namespace_id=namespace_id, group_name=group_name,
                                          cluster_name=cluster_name, healthy=healthy, weight=weight, enabled=enabled, metadata=metadata, ephemeral=ephemeral)
            if result.content == b'ok':
                logger.info(
                    f'[InstanceRegister] register success! service_name: {service_name}, ip: {ip}, port: {port}, group_name: {group_name}, namespace_id: {namespace_id}')
                while True:
                    # 发送心跳
                    beat_data = {
                        "serviceName": f'{group_name}@@{service_name}',
                        "ip": ip,
                        "port": port,
                        "weight": weight,
                        "groupName": group_name,
                        "clusterName": cluster_name,
                        "ephemeral": ephemeral,
                        "metadata": json.loads(metadata)
                    }
                    logger.info(
                        f'[InstanceRegister] send heartbeat. service_name: {service_name}, ip: {ip}, port: {port}, group_name: {group_name}, namespace_id: {namespace_id}, beat: {beat_data}')
                    result = self.instance.send_heartbeat(service_name, ip, port, json.dumps(beat_data),
                                                          namespace_id=namespace_id, group_name=group_name, ephemeral=ephemeral)
                    if result.code != 10200:
                        logger.error(
                            f'[InstanceRegister] send heartbeat failed! service_name: {service_name}, ip: {ip}, port: {port}, group_name: {group_name}, namespace_id: {namespace_id}, beat: {beat_data}')
                        break
                    time.sleep(5)
            else:
                logger.error(
                    f'[InstanceRegister] register failed! service_name: {service_name}, ip: {ip}, port: {port}, group_name: {group_name}, namespace_id: {namespace_id}')
        except (KeyboardInterrupt, SystemExit, EOFError, OSError):
            pass
        except:
            logger.error(
                f'[InstanceRegister] register failed! service_name: {service_name}, ip: {ip}, port: {port}, group_name: {group_name}, namespace_id: {namespace_id}')

    def delete_instance(self, signum, frame):
        logger.info(f'[InstanceRegister] catch signal {signum}, frame: {frame}, delete instance!')
        if self.register_info:
            # 注销服务
            result = self.instance.delete(self.register_info['service_name'], self.register_info['ip'], self.register_info['port'], namespace_id=self.register_info['namespace_id'],
                                          group_name=self.register_info['group_name'], ephemeral=self.register_info['ephemeral'])
            if result.content == b'ok':
                logger.info(
                    f'[InstanceRegister] delete instance success! {self.register_info["service_name"]}, ip: {self.register_info["ip"]}, port: {self.register_info["port"]}, group_name: {self.register_info["group_name"]}, namespace_id: {self.register_info["namespace_id"]}')
        sys.exit(1)



import functools

import orjson
import pydantic
import pydantic.generics


def orjson_dumps(v, *, default):
    return orjson.dumps(v, default=default).decode()


class BaseModel(pydantic.BaseModel):
    class Config:
        json_loads = orjson.loads
        json_dumps = orjson_dumps
        underscore_attrs_are_private = True


parse_raw_as = functools.partial(pydantic.parse_raw_as, json_loads=orjson.loads)



class BaseNacosAPI:
    """Nacos API base class"""

    def __init__(self, client=None):
        self._client = client

    def _get(self, url, **kwargs):
        return self._client._do_request('GET', url, **kwargs)

    def _post(self, url, **kwargs):
        return self._client._do_request('POST', url, **kwargs)

    def _put(self, url, **kwargs):
        return self._client._do_request('PUT', url, **kwargs)

    def _delete(self, url, **kwargs):
        return self._client._do_request('DELETE', url, **kwargs)
